<div>
 <iframe src="{{ filename }}" id="{{ id }}" allowfullscreen="" width="100%" height="450px" frameborder="0">

{{ content }}

</iframe>


<script>
    console.log("couou");

    let iframe_{{ id }} = document.getElementById("{{ id }}");
    let scheme_button = document.querySelector("#button_0");

    scheme_button.addEventListener("click", function() {
        console.log("click")
    });
</script>
</div>